--
-- PS/2 serial port, input only
--
-- Version : 0242
--
-- Copyright (c) 2002 Daniel Wallner (jesus@opencores.org)
--
-- All rights reserved
--
-- Redistribution and use in source and synthezised forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
-- Redistributions of source code must retain the above copyright notice,
-- this list of conditions and the following disclaimer.
--
-- Redistributions in synthesized form must reproduce the above copyright
-- notice, this list of conditions and the following disclaimer in the
-- documentation and/or other materials provided with the distribution.
--
-- Neither the name of the author nor the names of other contributors may
-- be used to endorse or promote products derived from this software without
-- specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.
--
-- Please report bugs to the author, but before you do so, please
-- make sure that this is not a derivative work and that
-- you have the latest version of this file.
--
--
--
-- File history :
--
--      0242 : First release
-- extended key handling added by MIKEJ
--
-- modifying for CPC by Tobias Gubener
--

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use ieee.std_logic_unsigned.all;

entity CPC_ps2 is
	port(
		Clk							: in std_logic;
		Rst_n						: in std_logic;
		T1MHz						: in std_logic;
		PS2_Clk						: in std_logic;
		PS2_Data					: in std_logic;
		Joya						: in std_logic_vector(5 downto 0);
		Joyb						: in std_logic_vector(5 downto 0);
		Key_BCD						: in std_logic_vector(3 downto 0);
		Key_Data					: out std_logic_vector(7 downto 0);
		speed						: out std_logic;    --L=>Turbo  H=>Clone
		pal							: out std_logic;		--L=>50     H=>60
		resf12						: out std_logic		--reset
	);
end CPC_ps2;

architecture rtl of CPC_ps2 is

	signal	Key_Addr		: std_logic_vector(9 downto 0);
	signal	PS2_Sample		: std_logic;
	signal	PS2_Data_s		: std_logic;

	signal	RX_Bit_Cnt		: unsigned(3 downto 0);
	signal	RX_Byte			: unsigned(2 downto 0);
	signal	RX_ShiftReg		: std_logic_vector(7 downto 0);
	signal	RX_Release		: std_logic;
	signal	RX_Received		: std_logic;
	signal  RX_E0           : std_logic;

	signal	LookUp			: std_logic_vector(11 downto 0);

	signal	Matrix_Set		: std_logic;
	signal	Matrix_Clear	: std_logic;
	signal	Matrix_Wr_Addr	: unsigned(11 downto 0);

	type Matrix_Image is array (natural range <>) of std_logic_vector(7 downto 0);
	signal	Matrix			: Matrix_Image(0 to 10);
	signal	Joyai   		: std_logic_vector(7 downto 0);
	signal	Joybi    		: std_logic_vector(7 downto 0);
	signal  speedi			: std_logic;
	signal 	resetf			: std_logic_vector(3 downto 0);
	signal  keylock			: std_logic;
	signal	Tick1us			: std_logic;
	signal  T1MHzd			: std_logic;

begin
	process (Clk)
	begin
		if Clk'event and Clk = '1' then
			if resetf /= 0 OR (RX_Received = '1' and RX_ShiftReg = x"07") then	--F12	Reset
				resetf <= resetf - 1;
				resf12 <= '0';
			else
--				resetf = 0;
				resf12 <= '1';
			end if;	
		end if;
	end process;	

	speed <= not speedi;
	process (Clk, Rst_n)
		variable PS2_Data_r		: std_logic_vector(1 downto 0);
		variable PS2_Clk_r		: std_logic_vector(1 downto 0);
		variable PS2_Clk_State	: std_logic;
	begin
		if Rst_n = '0' then
			PS2_Sample <= '0';
			PS2_Data_s <= '0';
			PS2_Data_r := "11";
			PS2_Clk_r := "11";
			PS2_Clk_State := '1';
		elsif Clk'event and Clk = '1' then
			T1MHzd<=T1MHz;
			Tick1us <= T1MHz and NOT T1MHzd;
			if Tick1us = '1' then
				PS2_Sample <= '0';

				-- Deglitch
				if PS2_Data_r = "00" then
					PS2_Data_s <= '0';
				end if;
				if PS2_Data_r = "11" then
					PS2_Data_s <= '1';
				end if;
				if PS2_Clk_r = "00" then
					if PS2_Clk_State = '1' then
						PS2_Sample <= '1';
					end if;
					PS2_Clk_State := '0';
				end if;
				if PS2_Clk_r = "11" then
					PS2_Clk_State := '1';
				end if;

				-- Double synchronise
				PS2_Data_r(1) := PS2_Data_r(0);
				PS2_Clk_r(1) := PS2_Clk_r(0);
				PS2_Data_r(0) := PS2_Data;
				PS2_Clk_r(0) := PS2_Clk;
			end if;
		end if;
	end process;

	process (Clk, Rst_n)
		variable Cnt : integer;
	begin
		if Rst_n = '0' then
			RX_Bit_Cnt <= (others => '0');
			RX_ShiftReg <= (others => '0');
			RX_Received <= '0';
			Cnt := 0;
		elsif Clk'event and Clk = '1' then
			RX_Received <= '0';
			if Tick1us = '1' then

				if PS2_Sample = '1' then
					if RX_Bit_Cnt = "0000" then
						if PS2_Data_s = '0' then -- Start bit
							RX_Bit_Cnt <= RX_Bit_Cnt + 1;
						end if;
					elsif RX_Bit_Cnt = "1001" then -- Parity bit
						RX_Bit_Cnt <= RX_Bit_Cnt + 1;
						-- Ignoring parity
					elsif RX_Bit_Cnt = "1010" then -- Stop bit
						if PS2_Data_s = '1' then
							RX_Received <= '1';
						end if;
						RX_Bit_Cnt <= "0000";
					else
						RX_Bit_Cnt <= RX_Bit_Cnt + 1;
						RX_ShiftReg(6 downto 0) <= RX_ShiftReg(7 downto 1);
						RX_ShiftReg(7) <= PS2_Data_s;
					end if;
				end if;

				-- TimeOut
				if PS2_Sample = '1' then
					Cnt := 0;
				elsif Cnt = 127 then
					RX_Bit_Cnt <= "0000";
					Cnt := 0;
				else
					Cnt := Cnt + 1;
				end if;
			end if;
		end if;
	end process;

	process (Clk, Rst_n)
	begin
		if Rst_n = '0' then
			RX_Byte <= (others => '0');
			RX_Release <= '0';
			Matrix_Set <= '0';
			Matrix_Clear <= '0';
			Matrix_Wr_Addr <= (others => '0');
			RX_E0 <= '0';
			keylock <= '0';
		elsif Clk'event and Clk = '1' then
			Matrix_Set <= '0';
			Matrix_Clear <= '0';

			if RX_Received = '1' then
				RX_Byte <= RX_Byte + 1;
				if RX_ShiftReg = x"F0" then
					RX_Release <= '1';
				elsif RX_ShiftReg = x"E0" then
					RX_E0 <= '1';
				elsif RX_ShiftReg = x"01" then 	--F9   	Clone
					speedi <= '0';
				elsif RX_ShiftReg = x"09" then	--F10	Turbo
					speedi <= '1';
				elsif RX_ShiftReg = x"03" then 	--F5   	PAL
					pal <= '0';
				elsif RX_ShiftReg = x"0B" then	--F6	NTSC
					pal <= '1';
				elsif RX_ShiftReg = x"0A" then	--F8	Keylock
					keylock <= '1';
				elsif (RX_ShiftReg = x"76" or RX_ShiftReg = x"5A") and keylock='1' then	--ESC or Enter	clr Keylock
					keylock <= '0';
				else
					RX_E0 <= '0';
					RX_Release <= '0';
					-- Normal key press
					if unsigned(LookUp) /= 0 and RX_Release = '0' then
						Matrix_Wr_Addr <= unsigned(LookUp);
--						Matrix_Set <= '1';
						Matrix_Set <= not keylock;
					end if;
					-- Normal key release
					if unsigned(LookUp) /= 0 and RX_Release = '1' then
						Matrix_Wr_Addr <= unsigned(LookUp);
						Matrix_Clear <= '1';
					end if;
				end if;
			end if;
		end if;
	end process;
	

	process (RX_ShiftReg, RX_E0)
	begin
	if (RX_E0 = '1')THEN
		case RX_ShiftReg is
		when x"71" => LookUp <= "001000000001"; -- CLR 
		when x"14" => LookUp <= "101010000000"; -- CTRL
		when x"5a" => LookUp <= "000001000000"; -- return
		when x"11" => LookUp <= "000100000010"; -- COPY
		when x"6b" => LookUp <= "000100000001"; -- left
		when x"75" => LookUp <= "000000000001"; -- up
		when x"74" => LookUp <= "000000000010"; -- right
		when x"72" => LookUp <= "000000000100"; -- down
		when others => LookUp <= "000000000000";
		end case;

	else
		case RX_ShiftReg is
		when x"12" => LookUp <= "001000100000"; -- left shift
		when x"1a" => LookUp <= "100010000000"; -- z
		when x"22" => LookUp <= "011110000000"; -- x
		when x"21" => LookUp <= "011101000000"; -- c
		when x"2a" => LookUp <= "011010000000"; -- v
		when x"1c" => LookUp <= "100000100000"; -- a
		when x"1b" => LookUp <= "011100010000"; -- s
		when x"23" => LookUp <= "011100100000"; -- d
		when x"2b" => LookUp <= "011000100000"; -- f
		when x"34" => LookUp <= "011000010000"; -- g
		when x"15" => LookUp <= "100000001000"; -- q
		when x"1d" => LookUp <= "011100001000"; -- w
		when x"24" => LookUp <= "011100000100"; -- e
		when x"2d" => LookUp <= "011000000100"; -- r
		when x"2c" => LookUp <= "011000001000"; -- t
		when x"16" => LookUp <= "100000000001"; -- 1
		when x"69" => LookUp <= "000100100000"; -- 1
		when x"1e" => LookUp <= "100000000010"; -- 2
		when x"72" => LookUp <= "000101000000"; -- 2
		when x"26" => LookUp <= "011100000010"; -- 3
		when x"7a" => LookUp <= "000000100000"; -- 3
		when x"25" => LookUp <= "011100000001"; -- 4
		when x"6b" => LookUp <= "001000010000"; -- 4
		when x"2e" => LookUp <= "011000000010"; -- 5
		when x"73" => LookUp <= "000100010000"; -- 5
		when x"45" => LookUp <= "010000000001"; -- 0
		when x"70" => LookUp <= "000110000000"; -- 0
		when x"46" => LookUp <= "010000000010"; -- 9
		when x"7d" => LookUp <= "000000001000"; -- 9
		when x"3e" => LookUp <= "010100000001"; -- 8
		when x"75" => LookUp <= "000100001000"; -- 8
		when x"3d" => LookUp <= "010100000010"; -- 7
		when x"6c" => LookUp <= "000100000100"; -- 7
		when x"36" => LookUp <= "011000000001"; -- 6
		when x"74" => LookUp <= "000000010000"; -- 6
		when x"4d" => LookUp <= "001100001000"; -- p
		when x"44" => LookUp <= "010000000100"; -- o
		when x"43" => LookUp <= "010000001000"; -- i
		when x"3c" => LookUp <= "010100000100"; -- u
		when x"35" => LookUp <= "010100001000"; -- y
		when x"5a" => LookUp <= "001000000100"; -- return
		when x"4b" => LookUp <= "010000010000"; -- l
		when x"42" => LookUp <= "010000100000"; -- k
		when x"3b" => LookUp <= "010100100000"; -- j
		when x"33" => LookUp <= "010100010000"; -- h
		when x"29" => LookUp <= "010110000000"; -- Space
		when x"59" => LookUp <= "101000100000"; -- right shift
		when x"3a" => LookUp <= "010001000000"; -- m
		when x"31" => LookUp <= "010101000000"; -- n
		when x"32" => LookUp <= "011001000000"; -- b
		when x"76" => LookUp <= "100000000100"; -- ESC
		when x"0d" => LookUp <= "100000010000"; -- TAB
		when x"58" => LookUp <= "100001000000"; -- CAPS LOOK
		when x"71" => LookUp <= "000010000000"; -- NUM 
		when x"54" => LookUp <= "001000000010"; -- [
		when x"5b" => LookUp <= "001000001000"; -- ]
		when x"0e" => LookUp <= "001001000000"; -- backslash
		when x"14" => LookUp <= "001010000000"; -- CTRL
		when x"55" => LookUp <= "001100000001"; -- ^
		when x"4e" => LookUp <= "001100000010"; -- =
		when x"5d" => LookUp <= "001100000100"; -- @
		when x"61" => LookUp <= "001100000100"; -- @
		when x"4c" => LookUp <= "001100010000"; -- ;
		when x"52" => LookUp <= "001100100000"; -- :
		when x"4a" => LookUp <= "001101000000"; -- /
		when x"49" => LookUp <= "001110000000"; -- .
		when x"41" => LookUp <= "010010000000"; -- ,
		when x"66" => LookUp <= "100110000000"; -- DEL
		when others => LookUp <= "000000000000";
		end case;
	end if;
	end process;

	process (Clk, Rst_n)
	begin
		if Rst_n = '0' then
			Matrix <= (others => (others => '0'));
		elsif Clk'event and Clk = '1' then
--			if RX_ShiftReg = x"aa" and RX_Received = '1' then
--				Matrix <= (others => (others => '0'));
--			end if;
			if Matrix_Set = '1' then
				Matrix(to_integer(Matrix_Wr_Addr(11 downto 8))) <=
					Matrix(to_integer(Matrix_Wr_Addr(11 downto 8))) or
					std_logic_vector(Matrix_Wr_Addr(7 downto 0));
			end if;
			if Matrix_Clear = '1' then
				Matrix(to_integer(Matrix_Wr_Addr(11 downto 8))) <=
					Matrix(to_integer(Matrix_Wr_Addr(11 downto 8))) and
					std_logic_vector(not Matrix_Wr_Addr(7 downto 0));
			end if;
		end if;
	end process;
	
	process (Key_BCD)
	begin
		case Key_BCD is
 		when	"1001"	=> Key_Addr <= "0111111111";
		when	"1000"	=> Key_Addr <= "1011111111";
		when	"0111"	=> Key_Addr <= "1101111111";
		when	"0110"	=> Key_Addr <= "1110111111";
		when	"0101"	=> Key_Addr <= "1111011111";
		when	"0100"	=> Key_Addr <= "1111101111";
		when	"0011"	=> Key_Addr <= "1111110111";
		when	"0010"	=> Key_Addr <= "1111111011";
		when	"0001"	=> Key_Addr <= "1111111101";
		when	"0000"	=> Key_Addr <= "1111111110";
		when others	=> Key_Addr <= "1111111111";
		end case;
	end process;
	

	Joyai <= ('0'&'0'&Joya);
	Joybi <= ('0'&'0'&Joyb);
	
	g_out1 : for i in 0 to 7 generate
		Key_Data(i) <= not ((Matrix(0)(i) and not Key_Addr(0)) or
					(Matrix(1)(i) and not Key_Addr(1)) or
					((Matrix(2)(i) or Matrix(10)(i)) and not Key_Addr(2)) or
--					(Matrix(2)(i) and not Key_Addr(2)) or
					(Matrix(3)(i) and not Key_Addr(3)) or
					(Matrix(4)(i) and not Key_Addr(4)) or
					(Matrix(5)(i) and not Key_Addr(5)) or
					((Matrix(6)(i) or Joyai(i)) and not Key_Addr(6)) or
--					(Matrix(6)(i) and not Key_Addr(6)) or
					(Matrix(7)(i) and not Key_Addr(7)) or
					(Matrix(8)(i) and not Key_Addr(8)) or
					((Matrix(9)(i) or Joybi(i)) and not Key_Addr(9)));
--					(Matrix(9)(i) and not Key_Addr(9)));
	end generate;
end;
